package cs108;

import static java.nio.charset.StandardCharsets.UTF_8;
import static java.util.Collections.max;
import static java.util.Comparator.comparingInt;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.function.ToIntFunction;

public final class Main {
    public static void main(String[] args) throws IOException {
        Collection<String> words = loadWords("mots_1.txt");
        Set<String> javaCollisions = largestCollision(words, String::hashCode);
        Set<String> sipHashCollisions = largestCollision(words, Main::sipHashString);
        System.out.printf("Collisions hashCode    : %3d (%s)%n", javaCollisions.size(), javaCollisions);
        System.out.printf("Collisions SipHash-2-4 : %3d (%s)%n", sipHashCollisions.size(), sipHashCollisions);
    }

    private static Set<String> largestCollision(Collection<String> words, ToIntFunction<String> hash) {
        Map<Integer, Set<String>> collisions = new HashMap<>(words.size());
        for (String w: words)
            collisions.computeIfAbsent(hash.applyAsInt(w), k -> new HashSet<>()).add(w);
        return max(collisions.values(), comparingInt(Set::size));
    }

    private static final byte[] KEY_BYTES = new byte[] { 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15 };
    private static final SipKey KEY = new SipKey(KEY_BYTES);
    private static final SipHasher_2_4 SIPHASHER_2_4 = new SipHasher_2_4(KEY);

    private static int sipHashString(String s) {
        return (int)SIPHASHER_2_4.hash(s.getBytes(UTF_8));
    }

    private static Collection<String> loadWords(String fileName) throws IOException {
        Collection<String> words = new ArrayList<>();
        try (BufferedReader r = new BufferedReader(
                new InputStreamReader(
                        new FileInputStream(fileName), UTF_8))) {
            String w;
            while ((w = r.readLine()) != null)
                words.add(w);
        }
        return words;
    }
}
