package cs108;

import static java.lang.Byte.toUnsignedLong;
import static java.lang.Long.rotateLeft;
import static java.nio.ByteOrder.LITTLE_ENDIAN;

import java.nio.ByteBuffer;

public final class SipHasher_2_4 {
    private final SipKey key;

    public SipHasher_2_4(SipKey key) {
        this.key = key;
    }

    public long hash(byte[] bytes) {
        State state = new State(key.k0(), key.k1());

        ByteBuffer byteB = ByteBuffer.wrap(bytes);
        byteB.order(LITTLE_ENDIAN);

        // Compression phase: compress all bytes except the last group
        while (byteB.remaining() >= Long.BYTES)
            state.sipCompress(byteB.getLong());

        // Compression phase: compress the last group, along with the length
        int remainingBytesCount = byteB.remaining();
        long lastBytes = (long)bytes.length << (8 * (7 - remainingBytesCount));
        for (int i = bytes.length - 1; i >= bytes.length - remainingBytesCount; --i)
            lastBytes = (lastBytes << 8) | toUnsignedLong(byteB.get(i));
        state.sipCompress(lastBytes);

        // Finalization phase.
        return state.sipFinalize();
    }

    private final static class State {
        private final static int C = 2, D = 4;
        private long v0, v1, v2, v3;

        public State(long k0, long k1) {
            v0 = k0 ^ 0x736f6d6570736575L;
            v1 = k1 ^ 0x646f72616e646f6dL;
            v2 = k0 ^ 0x6c7967656e657261L;
            v3 = k1 ^ 0x7465646279746573L;
        }

        public void sipCompress(long m) {
            v3 ^= m;
            for (int i = 0; i < C; ++i)
                sipRound();
            v0 ^= m;
        }

        public long sipFinalize() {
            v2 ^= 0xFF;
            for (int i = 0; i < D; ++i)
                sipRound();
            return v0 ^ v1 ^ v2 ^ v3;
        }

        private void sipRound() {
            v0 += v1;
            v2 += v3;
            v1 = rotateLeft(v1, 13);
            v3 = rotateLeft(v3, 16);
            v1 ^= v0;
            v3 ^= v2;
            v0 = rotateLeft(v0, 32);
            v2 += v1;
            v0 += v3;
            v1 = rotateLeft(v1, 17);
            v3 = rotateLeft(v3, 21);
            v1 ^= v2;
            v3 ^= v0;
            v2 = rotateLeft(v2, 32);
        }
    }
}
